/*
  +----------------------------------------------------------------------+
  | PHP Version 4                                                        |
  +----------------------------------------------------------------------+
  | Copyright (c) 1997-2002 The PHP Group                                |
  +----------------------------------------------------------------------+
  | This source file is subject to version 2.02 of the PHP license,      |
  | that is bundled with this package in the file LICENSE, and is        |
  | available at through the world-wide-web at                           |
  | http://www.php.net/license/2_02.txt.                                 |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Author: Xavier Noguer <xnoguer@php.net>                              |
  +----------------------------------------------------------------------+

  $Id: valkyrie.c,v 1.3 2004/04/08 23:46:50 xnoguer Exp $ 

  TODO: add an 
array valkyrie_errors(), function to access all errors found during validation
instead of dying on valkyrie_validate(). Will need to maintain a global array
of errors.

  TODO: check malloc's for global stuff for failure
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "zend_extensions.h" /* TODO: is this necessary? */
#include "php_ini.h"
#include "ext/standard/info.h"
#include "php_valkyrie.h"

ZEND_DECLARE_MODULE_GLOBALS(valkyrie)

/* True global resources - no need for thread safety here */
/*static int le_valkyrie;
*/

/* {{{ valkyrie_functions[]
 *
 * Every user visible function must have an entry in valkyrie_functions[].
 */
function_entry valkyrie_functions[] = {
	PHP_FE(valkyrie_validate,  NULL)
	PHP_FE(valkyrie_reload_config,  NULL)
	{NULL, NULL, NULL}	/* Must be the last line in valkyrie_functions[] */
};
/* }}} */


/* new code (from apd) */

// ---------------------------------------------------------------------------
// Zend Extension Support
// ---------------------------------------------------------------------------

int valkyrie_zend_startup(zend_extension *extension)
{
	TSRMLS_FETCH();
	CG(extended_info) = 1;  // XXX: this is ridiculous
	return zend_startup_module(&valkyrie_module_entry);
}

ZEND_DLEXPORT void valkyrie_zend_shutdown(zend_extension *extension)
{
	// Do nothing.
}

#ifndef ZEND_EXT_API
#define ZEND_EXT_API	ZEND_DLEXPORT
#endif
ZEND_EXTENSION();

ZEND_DLEXPORT zend_extension zend_extension_entry = {
	"Valkyrie",
	"0.1",
	"Xavier Noguer",
	"http://pear.php.net/",
	"",
	valkyrie_zend_startup,
	valkyrie_zend_shutdown,
	NULL,		// activate_func_t
	NULL,		// deactivate_func_t
	NULL,		// message_handler_func_t
	NULL,		// op_array_handler_func_t
	NULL,    // statement_handler_func_t
	NULL,	// fcall_begin_handler_func_t
	NULL,	// fcall_end_handler_func_t
	NULL,		// op_array_ctor_func_t
	NULL,		// op_array_dtor_func_t
#ifdef COMPAT_ZEND_EXTENSION_PROPERTIES
	NULL,		// api_no_check
	COMPAT_ZEND_EXTENSION_PROPERTIES
#else
	STANDARD_ZEND_EXTENSION_PROPERTIES
#endif
};

/* new code */



/* {{{ valkyrie_module_entry
 */
zend_module_entry valkyrie_module_entry = {
#if ZEND_MODULE_API_NO >= 20010901
	STANDARD_MODULE_HEADER,
#endif
	"valkyrie",
	valkyrie_functions,
	PHP_MINIT(valkyrie),
	PHP_MSHUTDOWN(valkyrie),
	PHP_RINIT(valkyrie),		/* Replace with NULL if there's nothing to do at request start */
	NULL,
	PHP_MINFO(valkyrie),
#if ZEND_MODULE_API_NO >= 20010901
	"0.1", /* Replace with version number for your extension */
#endif
	STANDARD_MODULE_PROPERTIES
};
/* }}} */

#ifdef COMPILE_DL_VALKYRIE
ZEND_GET_MODULE(valkyrie)
#endif

/* {{{ PHP_INI
 */
PHP_INI_BEGIN()
	STD_PHP_INI_ENTRY   ("valkyrie.config_path",  NULL, PHP_INI_ALL, OnUpdateString, config_path, zend_valkyrie_globals, valkyrie_globals)
	STD_PHP_INI_BOOLEAN ("valkyrie.auto_validate", "0", PHP_INI_ALL, OnUpdateInt, auto_validate, zend_valkyrie_globals, valkyrie_globals)
PHP_INI_END()
/* }}} */

/* {{{ php_valkyrie_init_globals
 */
static int php_valkyrie_init_globals(zend_valkyrie_globals *valkyrie_globals)
{
	valkyrie_globals->config_parsed = 0;
	//valkyrie_globals->validation_funcs_path = (char *)emalloc(1);
	//valkyrie_globals->doc_ptr = NULL;

	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MINIT_FUNCTION
 */
PHP_MINIT_FUNCTION(valkyrie)
{
	REGISTER_INI_ENTRIES();
	ZEND_INIT_MODULE_GLOBALS(valkyrie, php_valkyrie_init_globals, NULL);
	/* TODO: can't access php.ini values here, so we can't parse the config
       file. Find out why */
	VALKYRIE_G(zce) = NULL;
	VALKYRIE_G(application_path) = NULL;
	VALKYRIE_G(validation_funcs_path) = NULL;
	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MSHUTDOWN_FUNCTION
 */
PHP_MSHUTDOWN_FUNCTION(valkyrie)
{
	UNREGISTER_INI_ENTRIES();
	/* free the whole XML document (config file) */
	//xmlFreeDoc(VALKYRIE_G(doc_ptr));
	//xmlFreeNodeList(VALKYRIE_G(modules));
	//destroy_zend_class(VALKYRIE_G(zce));
	free(VALKYRIE_G(application_path));
	free(VALKYRIE_G(validation_funcs_path));
	return SUCCESS;
}
/* }}} */

/* {{{ get_elements_by_tagname, borrowed from domxml extension
 */
xmlNodeSetPtr get_elements_by_tagname(xmlNodePtr n, xmlChar* name, xmlNodeSet *rv )
{
	xmlNodePtr cld = NULL;
	/* TODO (from domxml extension!!)
	   Namespace support
	 */
	if ( n != NULL && name != NULL ) {
		cld = n->children;
		while ( cld != NULL ) {
			if ( xmlStrcmp( name, cld->name ) == 0 ){
				if ( rv == NULL ) {
					rv = xmlXPathNodeSetCreate( cld ) ;
				}
				else {
					xmlXPathNodeSetAdd( rv, cld );
				}
			}
			rv = get_elements_by_tagname(cld, name, rv);
			cld = cld->next;
		}
	}
	return rv;
}
/* }}} */

/* {{{ valkyrie_flag_param
   flag the received param in the GET or POST list */
void valkyrie_flag_param(valkyrie_param_to_receive **param_list,
                         char *param_name)
{
	valkyrie_param_to_receive *param_ptr;

	param_ptr = *param_list;
	while (param_ptr != NULL) {
		if (strcmp(param_ptr->name, param_name) == 0) {
			param_ptr->received = 1;
			break;
		}
		param_ptr = param_ptr->next;
	}
}
/* }}} */

/* {{{ valkyrie_get_input_params
   Get GET and POST params and create arrays for "strict" validation if needed
   Returns FAILURE or SUCCESS */
int valkyrie_get_input_params(zval ***get_data,
                              valkyrie_param_to_receive **get_received,
                              zval ***post_data,
                              valkyrie_param_to_receive **post_received
                              TSRMLS_DC)
{
	//zval *tmp;
	char *string_key;
	ulong num_key;
	int key_type, total_elements, i;
	valkyrie_param_to_receive *param_ptr = NULL;

	if (zend_hash_find(&EG(symbol_table), "_GET", sizeof("_GET"), (void **) get_data) != SUCCESS || Z_TYPE_PP(*get_data) != IS_ARRAY) {
		return FAILURE; /* _GET array not found */
	}
	if (zend_hash_find(&EG(symbol_table), "_POST", sizeof("_POST"), (void **) post_data) != SUCCESS || Z_TYPE_PP(*post_data) != IS_ARRAY) {
		return FAILURE; /* _POST array not found */
	}

	if (VALKYRIE_G(strict) == 1) {
		total_elements = zend_hash_num_elements(Z_ARRVAL_PP(*get_data));
		if (total_elements) {
		    zend_hash_internal_pointer_reset(Z_ARRVAL_PP(*get_data));
			for (i = 0; i < total_elements; i++, zend_hash_move_forward(Z_ARRVAL_PP(*get_data))) {
				key_type = zend_hash_get_current_key(Z_ARRVAL_PP(*get_data), &string_key, &num_key, 0);
				if (key_type == HASH_KEY_IS_LONG) {
					return FAILURE;/* something bad happened */
				}
				/* if it's the first param */
				if (param_ptr == NULL) {
					param_ptr = (valkyrie_param_to_receive *)emalloc(sizeof(valkyrie_param_to_receive));
					*get_received = param_ptr;
				}
				else {
					param_ptr->next = (valkyrie_param_to_receive *)emalloc(sizeof(valkyrie_param_to_receive));
					param_ptr = param_ptr->next;
				}
				param_ptr->name = (char *)emalloc(strlen(string_key) + 1);
				strcpy(param_ptr->name, string_key);
				param_ptr->received = 0;
				param_ptr->next = NULL;
			}
		}
		total_elements = zend_hash_num_elements(Z_ARRVAL_PP(*post_data));
		if (total_elements) {
			param_ptr = NULL; /* re-initialize pointer */
		    zend_hash_internal_pointer_reset(Z_ARRVAL_PP(*post_data));
			for (i = 0; i < total_elements; i++, zend_hash_move_forward(Z_ARRVAL_PP(*post_data))) {
				key_type = zend_hash_get_current_key(Z_ARRVAL_PP(*post_data), &string_key, &num_key, 0);
				if (key_type == HASH_KEY_IS_LONG) {
					return FAILURE;/* something bad happened */
				}
				/* if it's the first param */
				if (param_ptr == NULL) {
					param_ptr = (valkyrie_param_to_receive *)emalloc(sizeof(valkyrie_param_to_receive));
					*post_received = param_ptr;
				}
				else {
					param_ptr->next = (valkyrie_param_to_receive *)emalloc(sizeof(valkyrie_param_to_receive));
					param_ptr = param_ptr->next;
				}
				param_ptr->name = (char *)emalloc(strlen(string_key) + 1);
				strcpy(param_ptr->name, string_key);
				param_ptr->received = 0;
				param_ptr->next = NULL;
			}
		}
	}

	return SUCCESS;
}
/* }}} */

/* {{{ parse_config
 Parses the config file. Returns SUCCESS or FAILURE
 */
int parse_config(TSRMLS_D)
{
	xmlDoc *docp = NULL;
	xmlNode *root;
	char *path, *validationfuncs, *validationfuncs_real_path, *strict;
	//int ini_result;
	zend_op_array *dummy_op_array;
	zval *zval_validationfuncs_path;
	zend_class_entry *basura;
	zend_function tmp;

	/* nothing we can do without config_path */
	if (VALKYRIE_G(config_path) == NULL) {
		return FAILURE;
	}

	/* parse the config file */
	/*if (VALKYRIE_G(doc_ptr) != NULL) {
		xmlFreeDoc(VALKYRIE_G(doc_ptr));
	}*/
	docp = xmlParseFile(VALKYRIE_G(config_path));
	if (!docp) {
		return FAILURE; /* TODO: more friendly error */
	}
	/* save pointer for freeing mem at MSHUTDOWN */
	//VALKYRIE_G(doc_ptr) = docp;

	root = xmlDocGetRootElement(docp);
	if (!root) {
		return FAILURE; /* TODO: more friendly error */
	}
	if (strcmp(root->name,"app") != 0) {
		return FAILURE; /* TODO: more friendly error */
	}

	/* get the path for the app */
	path = xmlGetProp(root, "path");
	if (path == NULL) {
		return FAILURE; /* TODO: more friendly error */
	}
	if (strcmp(path,".") == 0) {
		VALKYRIE_G(application_path) = (char *)malloc(1);
		strcpy(VALKYRIE_G(application_path), "");
	}
	else {
		/* free VALKYRIE_G(application_path) if it existed */
		/*if (VALKYRIE_G(application_path) != NULL) {
			free(VALKYRIE_G(application_path));
		}*/
		/* FIXME: uncomment here for CLI testing */
		/*VALKYRIE_G(application_path) = (char *) malloc(strlen(path) + 1);
		strcpy(VALKYRIE_G(application_path), "");*/
		//VALKYRIE_G(application_path) = (char *) realloc(VALKYRIE_G(application_path), strlen(path) + 1);
		VALKYRIE_G(application_path) = (char *) malloc(1 + strlen(path) + 1);
		strcpy(VALKYRIE_G(application_path), "/");

		strcat(VALKYRIE_G(application_path), path);
	}
	/* FIXME: check VALKYRIE_G(application_path) for safe_mode */

	/* get the path for validationfuncs file */
	validationfuncs = xmlGetProp(root, "validationfuncs");
    if (validationfuncs == NULL) {
		return FAILURE; /* TODO: more friendly error */
	}
	validationfuncs_real_path = expand_filepath(validationfuncs, NULL TSRMLS_CC);
	/* free VALKYRIE_G(validation_funcs_path) if it existed */
	/*if (VALKYRIE_G(validation_funcs_path) != NULL) {
		free(VALKYRIE_G(validation_funcs_path));
	}*/
	VALKYRIE_G(validation_funcs_path) = (char *) malloc(strlen(validationfuncs_real_path) + 1);
	//VALKYRIE_G(validation_funcs_path) = (char *) realloc(VALKYRIE_G(validation_funcs_path), strlen(validationfuncs_real_path) + 1);
	strcpy(VALKYRIE_G(validation_funcs_path), validationfuncs_real_path);
	efree(validationfuncs_real_path);

	/* check the strict (check all received params) attribute */
	strict = xmlGetProp(root, "strict");
    if (strict == NULL) {
		VALKYRIE_G(strict) = 0; /* if not received it is assumed as "no" */
	}
	else {
		if (strcmp(strict, "yes") == 0) {
			VALKYRIE_G(strict) = 1;
		}
		else if (strcmp(strict, "no") == 0) {
			VALKYRIE_G(strict) = 0;
		}
		else {
			return FAILURE; /* unrecognized value */
		}
	}

	/* get all modules */
	VALKYRIE_G(modules) = get_elements_by_tagname(root, "module", NULL);
	if (VALKYRIE_G(modules) == NULL) {
		return FAILURE; /* TODO: more friendly error if there are no modules? */
	}

	/* compile the validations funcs file */
	MAKE_STD_ZVAL(zval_validationfuncs_path);
	ZVAL_STRING(zval_validationfuncs_path, VALKYRIE_G(validation_funcs_path), 1);
	dummy_op_array = compile_filename(ZEND_INCLUDE, zval_validationfuncs_path TSRMLS_CC);
	if (dummy_op_array == NULL) {
		return FAILURE;
	}
    destroy_op_array(dummy_op_array);
	efree(dummy_op_array);
	zval_dtor(zval_validationfuncs_path);
	FREE_ZVAL(zval_validationfuncs_path);

	if (zend_hash_find(EG(class_table), "valkyrie_validate", strlen("valkyrie_validate") + 1, (void **)&VALKYRIE_G(zce)) == FAILURE) {
		return FAILURE;
	}

	/* don't let RINIT parse again */
	VALKYRIE_G(config_parsed) = 1;

	zend_hash_init(&VALKYRIE_G(validation_methods), 50/*VALKYRIE_G(validation_methods).nNumOfElements*/, NULL, NULL, 1);
	zend_hash_copy(&VALKYRIE_G(validation_methods), &(VALKYRIE_G(zce)->function_table), NULL, &tmp, sizeof(zend_function));
	zend_hash_find(EG(class_table), "valkyrie_validate", strlen("valkyrie_validate") + 1, (void **)&basura);

	return SUCCESS;
}
/* }}} */

/* {{{ valkyrie_validate
   Validate GET and/or POST params for the currently executed script */
int valkyrie_validate(TSRMLS_D)
{
	int i, j;
	int result, file_found = 0;
	zval *zvalidate_func, *zretval;
	zval **post_data, **get_data, **zscript_name;//, **tmp;
	valkyrie_param_to_receive *get_received = NULL;
    valkyrie_param_to_receive *post_received = NULL;
	valkyrie_param_to_receive *aux_ptr;
	zval ***argv;
	char *script_name, *file, *app_and_module_path;
	xmlNodeSet *params;


	/* get SCRIPT_NAME instead of zend_get_executed_filename(), because we are
	   only interested in path starting from document root */
	if (zend_hash_find(&EG(symbol_table), "SCRIPT_NAME", sizeof("SCRIPT_NAME"), (void **) &zscript_name) == SUCCESS && Z_TYPE_PP(zscript_name) == IS_STRING) {
		script_name = Z_STRVAL_PP(zscript_name);
	}
	else {
		php_error(E_ERROR, "Could not determine script name");
		return 0;
		/* FIXME: for CLI testing */
		//script_name = "ext/valkyrie/prueba.php";
	}

	/* loop through the modules in order to find what module the current
	   script belongs to */
	for (i = 0; i < VALKYRIE_G(modules)->nodeNr; i++) {
		xmlNodePtr module_node = VALKYRIE_G(modules)->nodeTab[i];
		char *path, *module_path, *temp_path, *module_name;
		xmlNodeSet *files;


		/* get attributes from module element */
		module_name = xmlGetProp(module_node, "name");
		if (module_name == NULL) {
			php_error(E_ERROR, "Module element missing its name attribute\n");
			return 0;
		}
		path = xmlGetProp(module_node, "path");
		if (path == NULL) {
			php_error(E_ERROR, "Module element %s missing its path attribute\n", module_name);
			return 0;
		}

		/* initialize temp_path with app + current module path */
		if (strcmp(path,".") == 0) {
			module_path = (char *) emalloc(1);
			strcpy(module_path, "");
		}
		else {
			module_path = (char *) emalloc(1 + strlen(path) + 1);
			strcpy(module_path, path);
			/* TODO: use DEFAULT_SLASH */
			strcat(module_path, "/");
		}
		temp_path = (char *) emalloc(strlen(VALKYRIE_G(application_path)) + strlen(module_path) + 1 + 1);
		strcpy(temp_path, VALKYRIE_G(application_path));
		strcat(temp_path, "/"); /* TODO: use DEFAULT_SLASH */
		strcat(temp_path, module_path);

		/* what the name of the file should be according to SCRIPT_NAME */
		if (strlen(temp_path) >= strlen(script_name)) {
			continue;
		}
		file = estrdup(script_name + strlen(temp_path));

		/* compare application and module path to SCRIPT_NAME */
		app_and_module_path = (char *)estrndup(script_name, strlen(temp_path));
		if (strcmp(temp_path, app_and_module_path) == 0) {
			files = get_elements_by_tagname(module_node, "file", NULL);
			if (files != NULL) {
				/* loop through the files */
				for (j = 0; j < files->nodeNr; j++) {
					xmlNodePtr file_node = files->nodeTab[j];
					char *file_name;

					file_name = xmlGetProp(file_node, "name");
					if (file_name == NULL) {
						php_error(E_ERROR, "File element missing its name attribute for module %s\n", module_name);
						return 0;
					}

					/* compare file names from config file and SCRIPT_NAME */
					if (strcmp(file, file_name) == 0) {
						file_found = 1;
						params = get_elements_by_tagname(file_node, "param", NULL);
						/* free mem and leave modules/files loops */
						efree(module_path);
						efree(temp_path);
						break;
					}
				}
				if (file_found == 1) {
					break;
				}
			}
		}
		efree(app_and_module_path);
		efree(module_path);
		efree(temp_path);
	}
	if (file_found == 0) {
		php_error(E_ERROR, "File %s is not defined in the configuration file", script_name);
		return 0;
	}


	/* get REQUEST parameters (TODO: move this to get_input_params()) */
	/*if (zend_hash_find(&EG(symbol_table), "_POST", sizeof("_POST"), (void **) &post_data) != SUCCESS || Z_TYPE_PP(post_data) != IS_ARRAY) {
		php_error(E_ERROR, "_POST array not found");
		return 0;
	}
	if (zend_hash_find(&EG(symbol_table), "_GET", sizeof("_GET"), (void **) &get_data) != SUCCESS || Z_TYPE_PP(get_data) != IS_ARRAY) {
		php_error(E_ERROR, "_GET array not found");
		return 0;
	}*/

	if (valkyrie_get_input_params(&get_data, &get_received, &post_data, &post_received TSRMLS_CC) == FAILURE) {
		php_error(E_ERROR, "Error getting GET and POST params");
		return 0;
	}
	MAKE_STD_ZVAL(zvalidate_func);
	MAKE_STD_ZVAL(zretval);
	/* loop through params defined in config file */
	if (params != NULL) {
		for (i = 0; i < params->nodeNr; i++) {
			xmlNodePtr param_node = params->nodeTab[i];
			char *is_optional, *param_name, *param_type, *param_method;
			char *validate_func;
			zval **param_received;
 
			/* get attributes for param element */
			param_name = xmlGetProp(param_node, "name");
			if (param_name == NULL) {
				php_error(E_ERROR, "Param element missing its name attribute for file %s\n", file);
				return 0;
			}
			param_type = xmlGetProp(param_node, "type");
			if (param_type == NULL) {
				php_error(E_ERROR, "Param %s missing its type attribute for file %s\n", param_name, file);
				return 0;
			}
			param_method = xmlGetProp(param_node, "method");
			/* method attribute is obligatory. This may change in the future */
			/* if it changes get the _REQUEST array too */
			if (param_method == NULL) {
				php_error(E_ERROR, "Param %s missing its method attribute for file %s\n", param_name, file);
				return 0;
			}
			is_optional = xmlGetProp(param_node, "optional");

			/* find the parameter if it's post */
			/* TODO: more optimal search if "strict" is set */
			if (strcmp(param_method, "post") == 0) {
				if (zend_hash_find(Z_ARRVAL_PP(post_data), param_name, strlen(param_name) + 1, (void **) &param_received) != SUCCESS) {
					if ((is_optional == NULL) || (strcmp(is_optional, "no") == 0)) {
						php_error(E_ERROR, "Post param %s not received\n", param_name);
						return 0;
					}
					else { /* if it's optional continue */
						continue;
					}
				}
				/* if "strict" is set, flag as received */
				else if(VALKYRIE_G(strict) == 1) {
					valkyrie_flag_param(&post_received, param_name);
				}
			}
			/* find the parameter if it's get */
			else if (strcmp(param_method, "get") == 0) {
				if (zend_hash_find(Z_ARRVAL_PP(get_data), param_name, strlen(param_name) + 1, (void **) &param_received) != SUCCESS) {
					if ((is_optional == NULL) || (strcmp(is_optional, "no") == 0)) {
						php_error(E_ERROR, "Get param %s not received\n", param_name);
						return 0;
					}
					else { /* if it's optional continue */
						continue;
					}
				}
				/* if "strict" is set, flag as received */
				else if(VALKYRIE_G(strict) == 1) {
					valkyrie_flag_param(&get_received, param_name);
				}
			}
			/* TODO: search on _COOKIES too ? */
			/* param was not received */
			if (param_received == NULL) {
				php_error(E_ERROR, "Param %s not received\n", param_name);
				return 0;
			}
    
			validate_func = (char *)emalloc(8 + strlen(param_type) + 1);
			strcpy(validate_func, "validate");
			strcat(validate_func, param_type);
			strcat(validate_func, "\0");
    
			/* call validation function */
			ZVAL_STRING(zvalidate_func, validate_func, 1);
			ZVAL_FALSE(zretval);
			argv = emalloc(sizeof(zval **));
			argv[0] = &param_received[0];

			zend_class_entry *basura;
			zend_hash_find(EG(class_table), "valkyrie_validate", strlen("valkyrie_validate") + 1, (void **)&basura);
			result = call_user_function_ex(&(VALKYRIE_G(zce)->function_table), NULL, zvalidate_func, &zretval, 1, argv, 0, NULL TSRMLS_CC);
			//result = call_user_function_ex(&VALKYRIE_G(validation_methods), NULL, zvalidate_func, &zretval, 1, argv, 0, NULL TSRMLS_CC);
			if (result == FAILURE) {
                /* XXX: try reloading config before giving up. */
				/* this is probably due to a memory leak */
				parse_config(TSRMLS_C);
				zend_hash_find(EG(class_table), "valkyrie_validate", strlen("valkyrie_validate") + 1, (void **)&basura);
				result = call_user_function_ex(&(VALKYRIE_G(zce)->function_table), NULL, zvalidate_func, &zretval, 1, argv, 0, NULL TSRMLS_CC);
				if (result == FAILURE) {
					php_error(E_ERROR, "Could not call function Valkyrie_Validate::%s, %p\n", validate_func, basura);
					return 0;
				}
			}
    
			if (Z_TYPE_P(zretval) != IS_BOOL) {
				php_error(E_ERROR, "Invalid return type for Valkyrie_Validate::%s", validate_func);
				return 0;
			}
			if (Z_BVAL_P(zretval) == 0) {
				php_error(E_ERROR, "Invalid type for parameter %s", param_name);
				return 0;
			}
    
			efree(file);
			efree(validate_func);
			efree(argv);
		}
	}

	/* check all received params were declared if "strict" is set */
	if (VALKYRIE_G(strict) == 1) {
		while (post_received != NULL) {
			if (post_received->received == 0) {
				php_error(E_ERROR, "Received undeclared POST param %s\n", post_received->name);
				return 0;
			}
			aux_ptr = post_received;
			post_received = post_received->next;
			efree(aux_ptr->name);
			efree(aux_ptr);
		}
		while (get_received != NULL) {
			if (get_received->received == 0) {
				php_error(E_ERROR, "Received undeclared GET param %s\n", get_received->name);
				return 0;
			}
			aux_ptr = get_received;
			get_received = get_received->next;
			efree(aux_ptr->name);
			efree(aux_ptr);
		}
	}

	zval_dtor(zvalidate_func);
	FREE_ZVAL(zvalidate_func);
	zval_dtor(zretval);
	FREE_ZVAL(zretval);

	return 1;
}
/* }}} */

/* {{{ PHP_RINIT_FUNCTION
 */
PHP_RINIT_FUNCTION(valkyrie)
{
	int validated;

	/* FIXME: auto_validate doesn't work (problem with php.ini values) */
	/* if auto_validate is On call valkyrie_validate() on every request 
	   and parse_config() only if necessary */
	if (VALKYRIE_G(auto_validate) == 1) {
		if ((VALKYRIE_G(config_path) != NULL) && (VALKYRIE_G(config_parsed) == 0)) {
			if (parse_config(TSRMLS_C) == FAILURE) {
				return FAILURE;
			}
		}
		else {
			return FAILURE; /* nothing we can do without config_path */
		}
		validated = valkyrie_validate(TSRMLS_C);
		if (validated == 0) {
			return FAILURE; /* TODO: log failure?? */
		}
	}

	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MINFO_FUNCTION
 */
PHP_MINFO_FUNCTION(valkyrie)
{
	php_info_print_table_start();
	php_info_print_table_header(2, "valkyrie support", "enabled");
	php_info_print_table_end();

	DISPLAY_INI_ENTRIES();
}
/* }}} */

/* {{{ proto bool valkyrie_validate()
   Validate GET and/or POST params for the currently executed script */
PHP_FUNCTION(valkyrie_validate)
{
	int validated;

	/* parse config on demand */
	if (VALKYRIE_G(config_parsed) == 0) {
		if (parse_config(TSRMLS_C) == FAILURE) {
			php_error(E_ERROR, "Could not parse config file");
			RETURN_BOOL(0);
		}
	}
	validated = valkyrie_validate(TSRMLS_C);
	RETURN_BOOL(validated);
}
/* }}} */

/* {{{ proto bool valkyrie_reload_config()
   Reload configuration file */
PHP_FUNCTION(valkyrie_reload_config)
{
	if (parse_config(TSRMLS_C) == FAILURE) {
		RETURN_BOOL(0);
	}
	else {
		RETURN_BOOL(1);
	}
}
/* }}} */


/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
